unit BNWiz;

interface

{$IFDEF WIN32}
uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ComCtrls, ActnList;
{$ENDIF}

{$IFDEF LINUX}
uses
  SysUtils, Classes, QGraphics, QControls, QForms, QDialogs,
  QComCtrls, QActnList, QTypes;
{$ENDIF}

type

  { Forward declaration for WizActions and TBNWizSheet }
  TBNWizard = class;

  { Type declarations for Wizard Actions }
  { TBNWizAction - base action for other Wizard actions }
  TBNWizAction = class(TAction)
  private
    FWizard: TBNWizard;
    procedure SetWizard(const Value: TBNWizard);

  protected

  public
    destructor Destroy; override;

    function HandlesTarget(Target: TObject): boolean; override;
    procedure UpdateTarget(Target: TObject); override;

  published
    property Wizard: TBNWizard read FWizard write SetWizard;
  end;

  TBNWizBackAction = class(TBNWizAction)
  private

  protected

  public
    constructor Create(AOwner: TComponent); override;
    procedure ExecuteTarget(Target: TObject); override;
    procedure UpdateTarget(Target: TObject); override;
  published

  end;

  TBNWizNextAction = class(TBNWizAction)
  private

  protected

  public
    constructor Create(AOwner: TComponent); override;
    procedure ExecuteTarget(Target: TObject); override;
    procedure UpdateTarget(Target: TObject); override;
  published

  end;

  TBNWizFinishAction = class(TBNWizAction)
  private

  protected

  public
    constructor Create(AOwner: TComponent); override;
    procedure ExecuteTarget(Target: TObject); override;
    procedure UpdateTarget(Target: TObject); override;
  published

  end;

  TBNWizNextFinishAction = class(TBNWizAction)
  private
    FFinishCaption: TCaption;
    FNextCaption: TCaption;
    procedure SetFinishCaption(const Value: TCaption);
    procedure SetNextCaption(const Value: TCaption);

  protected

  public
    constructor Create(AOwner: TComponent); override;
    procedure ExecuteTarget(Target: TObject); override;
    procedure UpdateTarget(Target: TObject); override;
  published
    property FinishCaption: TCaption read FFinishCaption write SetFinishCaption;
    property NextCaption: TCaption read FNextCaption write SetNextCaption;
  end;

  TBNWizCancelAction = class(TBNWizAction)
  private

  protected

  public
    constructor Create(AOwner: TComponent); override;
    procedure ExecuteTarget(Target: TObject); override;
  published

  end;

  { TBNWizSheet - one "page" of the Wizard }

  // Declare a new event handler type
  TBNWizEvent = procedure (Sender: TObject; var CanDoNext, CanFinish: boolean)
                          of object;

  TBNWizSheet = class(TTabSheet)
  private

    { Wizard fields }
    FBNWizard: TBNWizard;
    FNextSheet: TBNWizSheet;
    FOnValidate: TBNWizEvent;

    { Wizard property Set procedures }
    procedure SetNextSheet(const Value: TBNWizSheet);
    procedure SetWizard(APageControl: TBNWizard);

  protected

  public

    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    { Wizard public methods }
    procedure Validate;

    { Wizard public properties }
    property Wizard: TBNWizard read FBNWizard write SetWizard;

  published

    { Wizard properties }
    property NextSheet: TBNWizSheet read FNextSheet write SetNextSheet;

    { Wizard event handlers }
    property OnValidate: TBNWizEvent read FOnValidate write FOnValidate;

  end;


  TBNWizard = class(TPageControl)
  private

    { Wizard fields }
    FCanDoNext: boolean;
    FCanFinish: boolean;
    FFirstSheet: TBNWizSheet;
    FOnNext: TNotifyEvent;
    FOnFinish: TNotifyEvent;
    FOnBack: TNotifyEvent;
    FOnCancel: TNotifyEvent;
    FActions: TList;


    { Wizard property Set procedures }
    procedure SetFirstSheet(const Value: TBNWizSheet);
    procedure SetOnCancel(const Value: TNotifyEvent);
    procedure SetOnFinish(const Value: TNotifyEvent);
    procedure SetOnNext(const Value: TNotifyEvent);
    procedure SetOnBack(const Value: TNotifyEvent);
    procedure SetCanDoNext(const Value: boolean);
    procedure SetCanFinish(const Value: boolean);
  protected
    property Actions: TList read FActions;

    procedure Loaded; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    { Wizard methods }
    function WizFindNextPage(CurPage: TBNWizSheet; GoForward: boolean): TBNWizSheet;
    procedure Back;
    procedure Next;
    procedure Finish;
    procedure Cancel;

    { Wizard properties}
    property CanDoNext: boolean read FCanDoNext write SetCanDoNext;
    property CanFinish: boolean read FCanFinish write SetCanFinish;

  published

    { Wizard properties }
    { The start sheet }
    property FirstSheet: TBNWizSheet read FFirstSheet write SetFirstSheet;

    { Surface Wizard Events }
    property OnBack: TNotifyEvent read FOnBack write SetOnBack;
    property OnNext: TNotifyEvent read FOnNext write SetOnNext;
    property OnFinish: TNotifyEvent read FOnFinish write SetOnFinish;
    property OnCancel: TNotifyEvent read FOnCancel write SetOnCancel;
  end;

implementation

{ TBNWizSheet }

constructor TBNWizSheet.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Self.TabVisible := False;         // Override default True setting
end;

destructor TBNWizSheet.Destroy;
var
  I: integer;
begin
  if FBNWizard <> nil then
  begin
    if Assigned(FBNWizard.FirstSheet) and (FBNWizard.FirstSheet = Self) then
    begin
      // Remove the first sheet reference
      FBNWizard.FirstSheet := nil;
    end;
    // Remove any NextSheet references
    for I := 0 to FBNWizard.PageCount - 1 do
    begin
      if (FBNWizard.Pages[I] is TBNWizSheet) and
         ((FBNWizard.Pages[I] as TBNWizSheet).NextSheet = Self) then
      begin
        (FBNWizard.Pages[I] as TBNWizSheet).NextSheet := nil;
      end;
    end;
  end;
  inherited Destroy;
end;

procedure TBNWizSheet.SetWizard(APageControl: TBNWizard);
var
  I: integer;
begin
  if FBNWizard <> APageControl then
  begin
    PageControl := APageControl;
    FBNWizard := APageControl;
    // If the FirstSheet isn't already assigned, and we were the first page to
    // be created, set the FirstSheet property to ourself
    if not Assigned(FBNWizard.FirstSheet) and (FBNWizard.Pages[0] = Self) then
    begin
      FBNWizard.FirstSheet := Self;
    end;
    // If the previous sheet hasn't got a NextSheet, set the NextSheet to
    // ourself
    I := FBNWizard.PageCount - 2;
    if I > -1 then
    begin
      (FBNWizard.Pages[I] as TBNWizSheet).NextSheet := Self;
    end;
  end;
end;

procedure TBNWizSheet.SetNextSheet(const Value: TBNWizSheet);
begin
  FNextSheet := Value;
end;

procedure TBNWizSheet.Validate;
var
  Next, Finish: boolean;
begin
  // Fire the OnValidate event.
  if Assigned(FOnValidate) and not (csDesigning in ComponentState) then
  begin
    // Pass the current values of CanDoNext & CanFinish to the event handler.
    if Assigned(Self.Wizard) then
    begin
      Next := Self.Wizard.CanDoNext;
      Finish := Self.Wizard.CanFinish;
    end
    else
    begin
      Next := False;
      Finish := False;
    end;
    // Call the event handler.
    FOnValidate(Self, Next, Finish);
    // Put changes back into the wizard.
    if Assigned(Self.Wizard) then
    begin
      Self.Wizard.CanDoNext := Next;
      Self.Wizard.CanFinish := Finish;
    end;
  end;

end;

{ TBNWizard }

constructor TBNWizard.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FActions := TList.Create;
end;

destructor TBNWizard.Destroy;
var
  I: integer;
begin
  // Clear any sheet references to things which are about to be destroyed
  for I := 0 to PageCount - 1 do
  begin
    TBNWizSheet(Pages[I]).NextSheet := nil;
    TBNWizSheet(Pages[I]).FBNWizard := nil;
  end;
  // Remove any references to self in any actions. Setting the Action's wizard
  // to nil will also delete the entry from the wizard's FActions list;
  while FActions.Count > 0 do
  begin
    TBNWizAction(FActions[0]).Wizard := nil;
  end;
  // Free the list of actions
  FreeAndNil(FActions);
  inherited Destroy;
end;

procedure TBNWizard.Loaded;
var
  I: integer;
begin
  inherited Loaded;
  for I := 0 to PageCount - 1 do
  begin
    if TObject(Pages[I]) is TBNWizSheet then
    begin
      TBNWizSheet(Pages[I]).Wizard := Self;
    end;
  end;
  if Assigned(FFirstSheet) then
  begin
    ActivePage := FFirstSheet;
    // Set visible state of Next/Finish button
    FFirstSheet.Validate;
  end;
end;

function TBNWizard.WizFindNextPage(CurPage: TBNWizSheet; GoForward: boolean): TBNWizSheet;
var
  I: integer;
begin
  if Assigned(CurPage) then
  begin
    if GoForward then
    begin
      // Going forward. If the next sheet property is set, return it, otherwise
      // return the current page.
      if Assigned(CurPage.NextSheet) then
      begin
        Result := CurPage.NextSheet;
        Exit;
      end;
    end
    else
    begin
      // Going backwards. Look through the pages until we find one whose NextSheet
      // property is the CurPage. If we don't find one, return the current page.
      for I := 0 to Self.PageCount - 1 do
      begin
        if TBNWizSheet(Pages[I]).NextSheet = CurPage then
        begin
          Result := Pages[I] as TBNWizSheet;
          Exit;
        end;
      end;
    end;
  end;
  // We get here if we don't find a next/prev sheet, or if TabVisible doesn't
  // match.
  Result := CurPage;
end;

procedure TBNWizard.Cancel;
begin
  // Fire the OnCancel event;
  if Assigned(FOnCancel) then
  begin
    FOnCancel(Self);
  end;
end;

procedure TBNWizard.Finish;
begin
  if CanFinish then
  begin
    // Fire the OnFinish event
    if Assigned(FOnFinish) then
    begin
      FOnFinish(Self);
    end;
  end;
end;

procedure TBNWizard.Next;
var
  Page: TBNWizSheet;
begin
  if CanDoNext or (csDesigning in ComponentState) then
  begin
    // Get the next page in the sequence
    Page := WizFindNextPage(ActivePage as TBNWizSheet, True);
    if (Page <> nil) and (Page <> ActivePage) and CanChange then
    begin
      // Fire the OnNext event
      if Assigned(FOnNext) and not (csDesigning in ComponentState) then
      begin
        FOnNext(Self);
      end;
      // Disable Next, as we don't yet know whether or not we can do it
      CanDoNext := False;
      // Call the Validate method for the page to see if we should enable
      // Next or Finish
      Page.Validate;
      // Display the next page
      ActivePage := Page;
    end;
  end;
end;

procedure TBNWizard.Back;
var
  Page: TBNWizSheet;
begin
  // Get the next page in the sequence
  Page := WizFindNextPage(ActivePage as TBNWizSheet, False);
  if (Page <> nil) and (Page <> ActivePage) and CanChange then
  begin
{    // As we're going back from something, we assume that Next should be enabled
    CanDoNext := True;}
    // Call the Validate method of the new page to see if Next or Finish should
    // be enabled
    Page.Validate;
    // Display the previous page
    ActivePage := Page;
    // Fire the OnBack event
    if Assigned(FOnBack) and not (csDesigning in ComponentState) then
    begin
      FOnBack(Self);
    end;
  end;
end;

procedure TBNWizard.SetFirstSheet(const Value: TBNWizSheet);
begin
  FFirstSheet := Value;
  if ComponentState * [csLoading, csDesigning] = [] then
  begin
    ActivePage := Value;
  end;
end;

procedure TBNWizard.SetOnCancel(const Value: TNotifyEvent);
begin
  FOnCancel := Value;
end;

procedure TBNWizard.SetOnFinish(const Value: TNotifyEvent);
begin
  FOnFinish := Value;
end;

procedure TBNWizard.SetOnNext(const Value: TNotifyEvent);
begin
  FOnNext := Value;
end;

procedure TBNWizard.SetOnBack(const Value: TNotifyEvent);
begin
  FOnBack := Value;
end;

procedure TBNWizard.SetCanDoNext(const Value: boolean);
begin
  FCanDoNext := Value;
end;

procedure TBNWizard.SetCanFinish(const Value: boolean);
begin
  FCanFinish := Value;
end;

{ TBNWizAction }

destructor TBNWizAction.Destroy;
begin
  // If we've assigned a Wizard, we have to remove ourselves from the Wizard's
  // list of actions
  if Assigned(FWizard) then
  begin
    Wizard := nil;        // SetWizard will take care of the list of actions
  end;
  inherited;
end;

function TBNWizAction.HandlesTarget(Target: TObject): boolean;
begin
  Result := Target is TControl;         // Handle virtually anything
end;

procedure TBNWizAction.SetWizard(const Value: TBNWizard);
var
  I: integer;
begin
  if FWizard <> Value then
  begin
    // The wizard needs to keep a list of any actions linked to it, so it can
    // remove the links when it is destroyed. If we don't do this, we get
    // problems with the Form Designer. Note that FreeNotification/Notification
    // doesn't work in design mode.
    // Remove existing link
    if Assigned(FWizard) then
    begin
      I := FWizard.Actions.IndexOf(Self);
      if I >= 0 then
      begin
        FWizard.Actions.Delete(I);
      end;
    end;
    FWizard := Value;
    // Add new link
    if Assigned(Value) then
    begin
      Value.Actions.Add(Self);
    end;
  end;
end;

procedure TBNWizAction.UpdateTarget(Target: TObject);
begin
  inherited;
//  Enabled := Assigned(Wizard);          // Test moved to stop flickering
end;

{ TBNWizBackAction }

constructor TBNWizBackAction.Create(AOwner: TComponent);
begin
  inherited;
  if (csDesigning in ComponentState) then
  begin
    Caption := '< &Back';
  end;
end;

procedure TBNWizBackAction.ExecuteTarget(Target: TObject);
begin
  if Assigned(Wizard) then
  begin
    // Move to the previous page
    Wizard.Back;
  end;
  inherited;

end;

procedure TBNWizBackAction.UpdateTarget(Target: TObject);
begin
  inherited;
  if Assigned(Wizard) then
  begin
    // If we're back at the first page, disable the Back button
//    Enabled := Enabled and (Wizard.ActivePage <> Wizard.FirstSheet);
{    if (Enabled <> Enabled and
            (Wizard.WizFindNextPage(Wizard.ActivePage as TBNWizSheet, False) <>
             Wizard.ActivePage)) then
    begin}
//      Enabled := Enabled and
      Enabled := Assigned(Wizard) and
              (Wizard.WizFindNextPage(Wizard.ActivePage as TBNWizSheet, False) <>
               Wizard.ActivePage);
//    end;
  end;
end;

{ TBNWizNextAction }

constructor TBNWizNextAction.Create(AOwner: TComponent);
begin
  inherited;
  if (csDesigning in ComponentState) then
  begin
    Caption := '&Next >';
  end;
end;

procedure TBNWizNextAction.ExecuteTarget(Target: TObject);
begin
  if Assigned(Wizard) then
  begin
    // Move to the next page
    Wizard.Next;
  end;
  inherited;

end;

procedure TBNWizNextAction.UpdateTarget(Target: TObject);
begin
  inherited;
  if Assigned(Wizard) then
  begin
{    // put in this test to stop flicker
    if Enabled <> Wizard.CanDoNext and
                  Assigned((Wizard.ActivePage as TBNWizSheet).NextSheet) then
    begin}
      // Mirror the wizard's CanDoNext property
//      Enabled := Wizard.CanDoNext;              // Test moved to stop flicker
      // If we're at the last page, disable the Next button
//      Enabled := Enabled and
      Enabled := Wizard.CanDoNext and
                 Assigned((Wizard.ActivePage as TBNWizSheet).NextSheet);
//    end;
  end;
end;

{ TBNWizFinishAction }

constructor TBNWizFinishAction.Create(AOwner: TComponent);
begin
  inherited;
  if (csDesigning in ComponentState) then
  begin
    Caption := '&Finish';
  end;
end;

procedure TBNWizFinishAction.ExecuteTarget(Target: TObject);
begin
  if Assigned(Wizard) then
  begin
    // Finish
    Wizard.Finish;
  end;
  inherited;

end;

procedure TBNWizFinishAction.UpdateTarget(Target: TObject);
begin
  inherited;
{  // test enabled property to stop flicker
  if Assigned(Wizard) and (Enabled <> Wizard.CanFinish) then
  begin}
    Enabled := Wizard.CanFinish;
//  end;
end;

{ TBNWizCancelAction }

constructor TBNWizCancelAction.Create(AOwner: TComponent);
begin
  inherited;
  if (csDesigning in ComponentState) then
  begin
    Caption := '&Cancel';
  end;
end;

procedure TBNWizCancelAction.ExecuteTarget(Target: TObject);
begin
  if Assigned(Wizard) then
  begin
    // Abandon the wizard
    Wizard.Cancel;
  end;
  inherited;

end;

{ TBNWizNextFinishAction }

constructor TBNWizNextFinishAction.Create(AOwner: TComponent);
begin
  inherited;
  if (csDesigning in ComponentState) then
  begin
    // Default values for captions
    NextCaption := '&Next >';           // Will also set Caption property
    FinishCaption := '&Finish';
  end;
end;

procedure TBNWizNextFinishAction.ExecuteTarget(Target: TObject);
begin
  if Assigned(Wizard) then
  begin
    if not Wizard.CanFinish then
    begin
      // Move to the next page
      Wizard.Next;
    end
    else
    begin
      Wizard.Finish;
    end;
  end;
  inherited;

end;

procedure TBNWizNextFinishAction.SetFinishCaption(const Value: TCaption);
begin
  FFinishCaption := Value;
end;

procedure TBNWizNextFinishAction.SetNextCaption(const Value: TCaption);
begin
  // Keep the NextCaption & the Caption in sync if we're designing
  if (csDesigning in ComponentState) and (Caption = FNextCaption) then
  begin
    Caption := Value;
  end;
  FNextCaption := Value;
end;

procedure TBNWizNextFinishAction.UpdateTarget(Target: TObject);
begin
  inherited;
  if Assigned(Wizard) then
  begin
    // If we're at the last page, behave like a Finish button, else behave like
    // a Next button
    if not Assigned((Wizard.ActivePage as TBNWizSheet).NextSheet) then
    begin
      // Behave like a Finish button. Change to the Finish caption.
      if (FinishCaption <> '') then
      begin
        Caption := FinishCaption;
      end;
      if Enabled <> Wizard.CanFinish then         // put in this test to stop flicker
      begin
        Enabled := Wizard.CanFinish;
      end;
    end
    else
    begin
      // Behave like a Next button. Change to the Next caption.
      if (NextCaption <> '') then
      begin
        Caption := NextCaption;
      end;
      if Enabled <> Wizard.CanDoNext then         // put in this test to stop flicker
      begin
        Enabled := Wizard.CanDoNext;
      end;
    end;
  end;
end;

end.
